# from https://github.com/hyperopt/hyperopt/
# pip install version has error on python3 then temporary included

from __future__ import absolute_import

from .base import STATUS_STRINGS
from .base import STATUS_NEW
from .base import STATUS_RUNNING
from .base import STATUS_SUSPENDED
from .base import STATUS_OK
from .base import STATUS_FAIL

from .base import JOB_STATES
from .base import JOB_STATE_NEW
from .base import JOB_STATE_RUNNING
from .base import JOB_STATE_DONE
from .base import JOB_STATE_ERROR

from .base import Ctrl
from .base import Trials
from .base import trials_from_docs
from .base import Domain

from .fmin import fmin
from .fmin import fmin_pass_expr_memo_ctrl
from .fmin import FMinIter
from .fmin import partial
from .fmin import space_eval

# -- syntactic sugar
from . import hp

# -- exceptions
from . import exceptions

# -- Import built-in optimization algorithms
from . import rand
from . import tpe
from . import mix
from . import anneal

__version__ = '0.2'
